unit Univpcx;
{
  PCX file loader v1.0b by Daniel Parnell
  for the Universal Bitmap Loader v1.0b

  Mail to:
    s921878@minyos.its.rmit.edu.au

  Or WEB page:
    http://minyos.its.rmit.edu.au/~s921878
}

interface
uses WinProcs,WinTypes,Graphics,SysUtils,Univbm,Classes;

type
  TUnivPCX = class(TComponent)
  private
    { Private declarations }
  protected
    { Protected declarations }
  public
    { Public declarations }
  published
    { Published declarations }
  end;

  TPCX=class(TBMLoader)
    procedure Load(name : string; bitmap : TBitmap); override;
  end;

  TPCX16Palette = array[0..15,1..3] of byte;

  PPCXHeader=^TPCXHeader;
  TPCXHeader=record
    Password  : byte;
    Version   : byte;
    Encoding  : byte;
    Bits      : byte;
    x1,y1,
    x2,y2     : word;
    hRez,vRes : word;
    palette   : TPCX16Palette;
    reserved  : byte;
    planes    : byte;
    bytes     : word;
    palDesc   : word;
    unused    : array[1..58] of byte;
  end;

procedure Register;

implementation

procedure TPCX.Load(name : string; bitmap : TBitmap);
var
  header     : PPCXHeader;
  bitbuffer  : PImageBuffer;
  lineBuf    : PImageBuffer;
  i,j        : word;
  size       : longint;
  wide,high  : longint;
  plane      : byte;
  addr       : word;
  left       : integer;
  line       : integer;
  rep,dat    : byte;

  function BitSet(b : word):boolean;
  begin
    if lineBuf^[b shr 3] and (1 shl (b and 7))<>0 then
      result:=TRUE
    else
      result:=FALSE;
  end;

  procedure SetBit(b : word; v : boolean);
  begin
    if v then
      bitBuffer^[b shr 3]:=bitBuffer^[b shr 3] or  ($01 shl (b and 7))
    else
      bitBuffer^[b shr 3]:=bitBuffer^[b shr 3] or  ($01 shl (b and 7))
                                               xor ($01 shl (b and 7));
  end;

begin
  FileOpen(name);

  bitBuffer:=nil;
  try
    new(header);

    blockRead(infile,header^,sizeof(TPCXHeader));

    with header^ do
      begin
        if password<>$0A then
          raise EBMException.Create(name+' is not a valid PCX file!');

        wide:=abs(x2-x1)+1;
        high:=abs(y2-y1)+1;

        Size:=Bytes*planes;

        if size>sizeof(TImageBuffer) then
          raise EBMException.Create(name+' is TOO big to load!');

        beginBitmap(wide,high);

        case bits*planes of
          1 : begin
                StartPalette(2);
                SetPalette(0,0,0,0);
                SetPalette(1,255,255,255);
                FinishPalette(2);
              end;
          4 : begin
                StartPalette(16);
                for i:=0 to 15 do
                  SetPalette(i,header^.palette[i,1],
                               header^.palette[i,2],
                               header^.palette[i,3]);
                FinishPalette(16);
              end;
          8 : begin
                StartPalette(256);
                FileSeek(FileSize(infile)-256*3);

                for i:=0 to 255 do
                  SetPalette(i,NextByte,NextByte,NextByte);
                FinishPalette(256);

                FileSeek(sizeof(TPCXHeader));
              end;
        end;

        addr:=0;
        plane:=0;
        left:=bytes;
        line:=0;

        GetMem(bitBuffer,size);
        fillChar(bitBuffer^,size,0);

        GetMem(lineBuf,bytes);

        repeat
          rep:=NextByte;
          if rep and $c0=$c0 then
            begin
              rep:=rep and $3f;
              dat:=NextByte;
            end
          else
            begin
              dat:=rep;
              rep:=1;
            end;

          for i:=1 to rep do
            begin
              lineBuf^[addr]:=dat;
              inc(addr);
              dec(left);
            end;

          if left<=0 then
            begin
              { Convert 16 colour PCX data into a forn that
                windows likes ie  from  rrrrrrrr  Plane 0
                                        gggggggg  Plane 1
                                        bbbbbbbb  Plane 2
                                        iiiiiiii  Plane 3

                                  to    rgbirbgi  Plane 0  }
              if bits*planes=4 then
                for i:=0 to bytes*8 do
                  SetBit(4*(i xor 1)+plane,BitSet(i xor 7))
              else
                if bits*planes=24 then  { Do the same for 24 bit images }
                  for i:=0 to bytes-1 do
                    bitBuffer^[i*3+(2-plane)]:=lineBuf^[i]
                 else
                  for i:=0 to bytes-1 do  { Mono and 8 bit images are OK }
                    bitBuffer^[i]:=lineBuf^[i];

              inc(plane);
              if plane>=planes then
                begin
                  plane:=0;

                  WriteLine(wide,line,bits*planes,bitBuffer);
                  inc(line);
                end;

              left:=bytes;
              addr:=0;
            end;
        until Line>=High;


        BitmapDone(bitmap,wide,high);

        CleanUp;
        FreeMem(LineBuf,bytes);
        FreeMem(BitBuffer,size);
      end;
  finally
    dispose(header);
    FileClose;
  end;
end;

procedure Register;
begin
  RegisterComponents('Bitmap', [TUnivPCX]);
end;

initialization
  TPCX.Create('PCX Format','PCX');
end.
